import zmq
import json
import os
import subprocess
import platform
from sqlalchemy import create_engine

__all__ = [
    "zmq_context",
    "status",
    "node_name",
    "node_description",
    "node_dependancies",
    "entropy_identity",
    "playbook",
    "databucket",
    "data_server",
    "save_dry_run",
    "dry_run_data",
    "dry_run_log",
    "terminate_node",
    "is_IPython",
]

status = None
node_name = ""
node_description = ""
node_dependancies = []
node_icon = ""
entropy_identity = None
playbook = None
databucket = None
data_server = None

zmq_context_variable = None
runtime_data = None
executor_input = None  #: address of executor zmq communication channel
executor_output = None

save_dry_run = False
dry_run_data = None
dry_run_log = None


def zmq_context():
    global zmq_context_variable
    if zmq_context_variable is None:
        zmq_context_variable = zmq.Context()
    return zmq_context_variable


def runtime_db_connect():
    global runtime_data
    SQLALCHEMY_DATABASE_URL = (
        f"postgresql://{data_server[0]}:{data_server[1]}"
        + f'@{data_server[2]}/{data_server[3]}?client_encoding="utf8"'
    )
    runtime_data = create_engine(SQLALCHEMY_DATABASE_URL).connect()


def runtime_db_close():
    runtime_data.close()


def is_IPython():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def terminate_node():
    if save_dry_run and runtime_data is None:
        with open("dry_run_data.json", "w") as outfile:
            json.dump(dry_run_data, outfile)
        dry_run_log.close()
        # save to entropyhub
        env = os.environ
        env["PYTHONPATH"] = os.getcwd()
        if platform.system() == "Windows":
            python_cmd = "python"
        else:
            python_cmd = "python3"
        subprocess.run(
            f"{python_cmd} -m entropyhub.submit_dry_run",
            env=env,
            shell=True,
            universal_newlines=True,
            start_new_session=True,
        )

    if is_IPython():

        class StopExecution(Exception):
            def _render_traceback_(self):
                pass

        raise StopExecution
    else:
        exit()



import os
from pathlib import Path

import sqlalchemy.engine
from alembic import command, script, op
from alembic.config import Config
from alembic.runtime import migration

from entropylab.pipeline.results_backend.sqlalchemy.project import param_store_file_path


class AlembicUtil:
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self._engine = engine

    @staticmethod
    def get_param_store_file_path():
        conn = op.get_bind()
        project_path = os.path.abspath(
            os.path.join(os.path.dirname(conn.engine.url.database), "..")
        )
        return param_store_file_path(project_path)

    @staticmethod
    def _abs_path_to(rel_path: str) -> str:
        source_path = Path(__file__).resolve()
        source_dir = source_path.parent
        return os.path.join(source_dir, rel_path)

    def upgrade(self) -> None:
        with self._engine.connect() as connection:
            alembic_cfg = self._alembic_build_config(connection)
            command.upgrade(alembic_cfg, "head")

    def stamp_head(self) -> None:
        with self._engine.connect() as connection:
            alembic_cfg = self._alembic_build_config(connection)
            command.stamp(alembic_cfg, "head")

    def db_is_up_to_date(self) -> bool:
        script_location = self._abs_path_to("")
        script_ = script.ScriptDirectory(script_location)
        with self._engine.begin() as conn:
            context = migration.MigrationContext.configure(conn)
            db_version = context.get_current_revision()
            latest_version = script_.get_current_head()
            return db_version == latest_version

    def _alembic_build_config(self, connection: sqlalchemy.engine.Connection) -> Config:
        config_location = self._abs_path_to("../alembic.ini")
        script_location = self._abs_path_to("")
        alembic_cfg = Config(config_location)
        alembic_cfg.set_main_option("script_location", script_location)
        """
        Modified by @urig to share a single engine across multiple (typically in-memory)
        connections, based on this cookbook recipe:
        https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
        """
        alembic_cfg.set_main_option("sqlalchemy.url", "sqlite:///:memory:")
        alembic_cfg.attributes["connection"] = connection  # overrides dummy url above
        return alembic_cfg



from __future__ import annotations

import contextlib
import copy
import hashlib
import os
import string
from pathlib import Path
from random import SystemRandom
from typing import Optional, Callable, List, Set

from filelock import FileLock
from tinydb import TinyDB, Query
from tinydb.storages import MemoryStorage
from tinydb.table import Table, Document

from entropylab.logger import logger
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.params.persistence.persistence import (
    Persistence,
    Commit,
)
from entropylab.pipeline.params.persistence.tinydb.storage import JSONPickleStorage

CURRENT_VERSION = "0.2"

INFO_DOC_ID = 1
INFO_TABLE = "info"
TEMP_DOC_ID = 1
TEMP_TABLE = "temp"
VERSION_KEY = "version"
REVISION_KEY = "revision"


class TinyDbPersistence(Persistence):
    def __init__(self, path: Optional[str] | Optional[Path] = None):
        if path is None:
            self.__is_in_memory_mode = True
            self.__db = TinyDB(storage=MemoryStorage)
            self.__filelock = contextlib.nullcontext()
            with self.__filelock:
                set_version(self.__db, CURRENT_VERSION)
        else:
            self.__is_in_memory_mode = False
            path = str(path)
            is_new = not os.path.isfile(path)
            self.__db = TinyDB(path, storage=JSONPickleStorage)
            Table.default_query_cache_capacity = 0
            self.__filelock = FileLock(path + ".lock")
            with self.__filelock:
                if is_new:
                    logger.debug(f"Creating new ParamStore JSON file at '{path}'")
                    set_version(self.__db, CURRENT_VERSION)
                else:
                    version = get_version(self.__db)
                    if version != CURRENT_VERSION:
                        raise EntropyError(
                            f"ParamStore JSON file at '{path}' is version {version}. "
                            f"Please upgrade to {CURRENT_VERSION}."
                        )

    def close(self):
        with self.__filelock:
            self.__db.close()

    def get_commit(
        self,
        commit_id: Optional[str] = None,
        commit_num: Optional[int] = None,
    ) -> Optional[Commit]:
        if commit_id is not None:
            doc = self.__get_commit_by_id(commit_id)
        elif commit_num is not None:
            doc = self.__get_commit_by_num(commit_num)
        else:
            doc = self.__get_latest_doc()
        return self.__doc_to_commit(doc)

    def __get_commit_by_id(self, commit_id: str) -> Document:
        with self.__filelock:
            result = self.__db.search(Query().metadata.id == commit_id)
            if len(result) == 0:
                raise EntropyError(f"Commit with id '{commit_id}' not found")
            if len(result) > 1:
                raise EntropyError(
                    f"{len(result)} commits with id '{commit_id}' found. "
                    f"Only one commit is allowed per id"
                )
            return result[0]

    def __get_commit_by_num(self, commit_num: int) -> Document:
        with self.__filelock:
            result = self.__db.get(doc_id=commit_num)
            if result is None:
                raise EntropyError(f"Commit with number '{commit_num}' not found")
            return result

    def get_latest_commit(self) -> Optional[Commit]:
        with self.__filelock:
            latest_doc = self.__get_latest_doc()
            if latest_doc:
                return self.__doc_to_commit(latest_doc)
            else:
                return None

    @staticmethod
    def __doc_to_commit(doc: Optional[Document]) -> Optional[Commit]:
        if not doc:
            return None
        else:
            return Commit(
                id=doc["metadata"]["id"],
                timestamp=doc["metadata"]["timestamp"],
                label=doc["metadata"]["label"],
                params=doc["params"],
                tags=doc["tags"],
            )

    def __get_latest_doc(self) -> Optional[Document]:
        with self.__filelock:
            docs = self.__db.all()
            return docs[-1] if docs else None

    def commit(
        self,
        commit: Commit,
        dirty_keys: Optional[Set[str]] = None,
    ) -> str:
        commit.id = self.__generate_commit_id()
        # TODO: Perhaps create the timestamp here?
        self.stamp_dirty_params_with_commit(commit, dirty_keys)
        doc = self.__build_document(commit)
        with self.__filelock:
            doc.doc_id = self.__next_doc_id()
            self.__db.insert(doc)
        return commit.id

    @staticmethod
    def __generate_commit_id() -> str:
        random_string = "".join(
            SystemRandom().choice(string.printable) for _ in range(32)
        ).encode("utf-8")
        return hashlib.sha1(random_string).hexdigest()

    def __build_document(self, commit: Commit) -> Document:
        """
        builds a document to be saved as a commit/temp in TinyDB.
        :param commit:
        :return: a dictionary describing the current state of the ParamStore
        """
        metadata = commit.to_metadata()
        if self.__is_in_memory_mode:
            params = copy.deepcopy(commit.params)
        else:
            params = commit.params
        return Document(
            dict(metadata=metadata.__dict__, params=params, tags=commit.tags),
            doc_id=0,
        )

    def __next_doc_id(self):
        with self.__filelock:
            last_doc = self.__get_latest_doc()
            if last_doc:
                return last_doc.doc_id + 1
            else:
                return 1

    def search_commits(
        self, label: Optional[str] = None, key: Optional[str] = None
    ) -> List[Commit]:
        def test_label(lbl: str) -> Callable:
            return lambda metadata_label: metadata_label == lbl if lbl else True

        def test_key(k: str) -> Callable:
            return lambda params: k in params if k else True

        with self.__filelock:
            docs = self.__db.search(
                Query().metadata.label.test(test_label(label))
                & Query().params.test(test_key(key))
            )
            return list(map(self.__doc_to_commit, docs))

    # noinspection PyShadowingNames
    def save_temp_commit(self, commit: Commit) -> None:
        with self.__filelock:
            table = self.__db.table(TEMP_TABLE)
            doc = self.__build_document(commit)
            doc.doc_id = TEMP_DOC_ID
            table.upsert(doc)

    """ Temporary State """

    def load_temp_commit(self) -> Commit:
        with self.__filelock:
            table = self.__db.table(TEMP_TABLE)
        doc = table.get(doc_id=1)
        if not doc:
            raise EntropyError(
                "Temp is empty. Use save_temp_commit() before using load_temp_commit()"
            )
        else:
            return self.__doc_to_commit(doc)


def set_version(db: TinyDB | str, version: str, revision: Optional[str] = ""):
    if isinstance(db, str):
        # TODO: Support locking
        db = TinyDB(db, storage=JSONPickleStorage)
    info_table = db.table(INFO_TABLE)
    info_doc = info_table.get(doc_id=INFO_DOC_ID)
    if info_doc:
        info_doc[VERSION_KEY] = version
        info_doc[REVISION_KEY] = revision
    else:
        info_table.insert({VERSION_KEY: version, REVISION_KEY: revision})


def get_version(db: TinyDB):
    if INFO_TABLE in db.tables():
        info_table = db.table(INFO_TABLE)
        info_doc = info_table.get(doc_id=INFO_DOC_ID)
        return info_doc[VERSION_KEY] if VERSION_KEY in info_doc else None
    return "0.1"


def check_version(path: str, old_version: str, new_version: str):
    if os.path.isfile(path):
        with TinyDB(path) as old_db:
            actual_version = get_version(old_db)
            if actual_version != old_version:
                raise EntropyError(
                    f"Cannot migrate file '{path}' from version {actual_version} to "
                    f"version {new_version}"
                )
    else:
        raise EntropyError(
            f"Cannot migrate file '{path}' to version {new_version} because the file "
            "does not exist"
        )



from __future__ import annotations

import os

import jsonpickle as jsonpickle
from tinydb import Storage

from entropylab.logger import logger


class JSONPickleStorage(Storage):
    def __init__(self, filename):
        self.filename = filename

    def read(self):
        if not os.path.isfile(self.filename):
            return None
        with open(self.filename) as handle:
            # noinspection PyBroadException
            try:
                s = handle.read()
                data = jsonpickle.decode(s)
                return data
            except BaseException:
                logger.exception(
                    f"Exception decoding TinyDB JSON file '{self.filename}'"
                )
                return None

    def write(self, data):
        # noinspection PyBroadException
        try:
            with open(self.filename, "w+") as handle:
                s = jsonpickle.encode(data)
                handle.write(s)
        except BaseException:
            logger.exception(f"Exception encoding TinyDB JSON file '{self.filename}'")

    def close(self):
        pass



from __future__ import annotations

import copy
import threading
import time
from datetime import timedelta
from enum import unique, Enum
from pathlib import Path
from typing import Optional, Dict, Any, List, Set, MutableMapping

import pandas as pd

from entropylab.config import settings
from entropylab.pipeline.params.persistence.persistence import Commit, Metadata
from entropylab.pipeline.params.persistence.sqlalchemy.sqlalchemypersistence import (
    SqlAlchemyPersistence,
)
from entropylab.pipeline.params.persistence.tinydb.tinydbpersistence import (
    TinyDbPersistence,
)

UTC_TZ = "UTC"


@unique
class MergeStrategy(Enum):
    OURS = 1
    THEIRS = 2


# TODO: Convert to dataclass
class Param(Dict):
    def __init__(self, value):
        super().__init__()
        self.value: object = value
        self.commit_id: Optional[str] = None
        self.expiration: Optional[timedelta | pd.Timestamp] = None
        self.description: Optional[str] = None
        self.node_id: Optional[str] = None

    def __eq__(self, other):
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return (
            f"<Param(value={self.value}, "
            f"commit_id={self.commit_id}, "
            f"expiration={self.expiration})> "
        )

    @property
    def has_expired(self):
        """
        Indicates whether the Param value has expired. Returns True iff the Param
        has been committed and the time elapsed since the commit operation has exceeded
        the time recorded in the `expiration` property"""
        if isinstance(self.expiration, pd.Timestamp):
            return self.expiration < pd.Timestamp(time.time_ns())
        else:
            return False


class ParamStore(MutableMapping):
    """
    A class that provides versioned storage for experiment parameters (params).
    """

    def __init__(
        self,
        path: Optional[str] | Optional[Path] = None,
        url: Optional[str] | Optional[Path] = None,
        theirs: Optional[Dict | ParamStore] = None,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.THEIRS,
    ):
        super().__init__()
        self.__lock = threading.RLock()
        self.__params: Dict[str, Param] = dict()  # where current params are stored
        self.__tags: Dict[str, List[str]] = dict()  # tags that are mapped to keys
        self.__dirty_keys: Set[str] = set()  # updated keys not committed yet
        # constructor arguments take precedence:
        if path:
            self.__persistence = TinyDbPersistence(path)
        elif url:
            self.__persistence = SqlAlchemyPersistence(url)
        else:
            # ...over configuration settings
            if "param_store_path" in settings:
                self.__persistence = TinyDbPersistence(settings.param_store_path)
            elif "param_store_url" in settings:
                self.__persistence = SqlAlchemyPersistence(settings.param_store_url)
            else:
                # default
                self.__persistence = TinyDbPersistence()

        self.checkout()
        if theirs is not None:
            self.merge(theirs, merge_strategy)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__persistence.close()

    """ Properties """

    @property
    def is_dirty(self):
        """
        True iff params have been changed since the store has last been
        initialized or checked out.
        """
        pass
        with self.__lock:
            return len(self.__dirty_keys) > 0

    """ MutableMapping """

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set self[key] to value. The key-value pair becomes a "param" and
        can be persisted using `commit()` and retrieved later using
        `checkout()`.

        Note: Keys should not start with a dunder (`__`). Such keys are not
        treated as params and are not persisted when `commit()` is called.
        """

        if key.startswith("__") or key.startswith(f"_{self.__class__.__name__}__"):
            # keys that are private attributes are not params and are treated
            # as regular object attributes
            object.__setattr__(self, key, value)
        else:
            with self.__lock:
                self.__params.__setitem__(key, Param(value))
                self.__dirty_keys.add(key)

    def __getitem__(self, key: str) -> Any:
        with self.__lock:
            return self.__params.__getitem__(key).value

    def __delitem__(self, *args, **kwargs):
        with self.__lock:
            key = args[0]
            self.__params.__delitem__(*args, **kwargs)
            self.__remove_key_from_tags(key)
            self.__dirty_keys.add(key)

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            try:
                return self[key]
            except KeyError:
                raise AttributeError(key)

    def __setattr__(self, key, value):
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            try:
                self[key] = value
            except BaseException:
                raise AttributeError(key)
        else:
            object.__setattr__(self, key, value)

    def __iter__(self):
        with self.__lock:
            values = _extract_param_values(self.__params)
            return values.__iter__()

    def __len__(self):
        with self.__lock:
            return self.__params.__len__()

    def __contains__(self, key):
        with self.__lock:
            return self.__params.__contains__(key)

    def __repr__(self):
        with self.__lock:
            return f"<ParamStore({self.to_dict().__repr__()})>"

    """ Params """

    def keys(self):
        with self.__lock:
            return self.__params.keys()

    def to_dict(self) -> Dict:
        with self.__lock:
            return _extract_param_values(self.__params)

    def get_value(self, key: str, commit_id: Optional[str] = None) -> object:
        """
        Returns the value of a param by its key

        :param key: the key identifying the param
        :param commit_id: an optional commit_id. If provided, the value will be
        returned from the specified commit
        """
        with self.__lock:
            if commit_id is None:
                return self[key]
            else:
                commit = self.__persistence.get_commit(commit_id)
                return commit.params[key].value

    def get_param(self, key: str, commit_id: Optional[str] = None) -> Param:
        """
        Returns a copy of the Param instance of a value stored in ParamStore

        :param key: the key identifying the param
        :param commit_id: an optional commit_id. If provided, the Param will be
        returned from the specified commit
        """
        with self.__lock:
            if commit_id is None:
                return copy.deepcopy(self.__params[key])
            else:
                commit = self.__persistence.get_commit(commit_id)
                return copy.deepcopy(commit.params[key])

    def set_param(self, key: str, value: object, **kwargs):
        """
        Sets a Param in the ParamStore

        :param key: the key identifying the param
        :param value: the value of the param
        :param key: an optional period of time (measured from the time the param
        is committed) after which the prime expires (i.e. param.has_expired
        returns True)
        """
        if "commit_id" in kwargs:
            raise ValueError("Setting commit_id in set_param() is not allowed")
        if "value" in kwargs:
            raise ValueError("Value can only be set through positional argument")
        with self.__lock:
            if key in self.__params:
                param = self.get_param(key)
            else:
                param = Param(value)
            param.value = value
            param.__dict__.update(kwargs)
            self.__params.__setitem__(key, param)
            self.__dirty_keys.add(key)

    def __remove_key_from_tags(self, key: str):
        for tag in self.__tags:
            if key in self.__tags[tag]:
                self.__tags[tag].remove(key)

    def rename_key(self, key: str, new_key: str):
        with self.__lock:
            if new_key in self.keys():
                raise KeyError(
                    f"Cannot rename key '{key}' to key that already exists: '{new_key}'"
                )
            self.__rename_key_in_tags(key, new_key)
            value = self.__getitem__(key)
            self.__setitem__(new_key, value)
            self.__delitem__(key)

    def __rename_key_in_tags(self, key, new_key):
        for item in self.__tags.items():
            tag = item[0]
            keys = item[1]
            if key in keys:
                self.__tags[tag].remove(key)
                self.__tags[tag].append(new_key)

    """ Commits """

    def commit(self, label: Optional[str] = None) -> str:
        with self.__lock:
            commit = Commit(
                label=label,
                params=self.__params,
                tags=self.__tags,
            )
            commit_id = self.__persistence.commit(commit, self.__dirty_keys)
            self.__dirty_keys.clear()
            return commit_id

    def checkout(
        self, commit_id: Optional[str] = None, commit_num: Optional[int] = None
    ) -> None:
        with self.__lock:
            commit = self.__persistence.get_commit(commit_id, commit_num)
            if commit:
                self.__checkout(commit)

    def __checkout(self, commit: Commit):
        self.__params.clear()
        self.__params.update(commit.params)
        self.__tags.clear()
        self.__tags.update(commit.tags)
        self.__dirty_keys.clear()

    def list_commits(self, label: Optional[str] = None) -> List[Metadata]:
        """
        Returns a list of commits

        :param label: an optional label, if given then only commits that match
        it will be returned
        """
        with self.__lock:
            commits = self.__persistence.search_commits(label)
            metadata = map(Commit.to_metadata, commits)
            return list(metadata)

    """ Merge """

    def merge(
        self,
        theirs: ParamStore,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.OURS,
    ) -> None:
        with self.__lock:
            ours = self
            self.__merge_trees(ours, theirs, merge_strategy)

    def __merge_trees(
        self,
        a: ParamStore | Dict,
        b: ParamStore | Dict,
        merge_strategy: MergeStrategy,
    ) -> bool:
        """Merges `b` into `a` *in-place* using the given strategy"""
        a_has_changed = False
        for key in b.keys():
            if key in a.keys():
                if (
                    (not isinstance(a[key], Param))
                    and isinstance(a[key], dict)
                    and (not isinstance(b[key], Param))
                    and isinstance(b[key], dict)
                ):
                    """This is a special case where the values of the Params are both
                    dictionaries. In this case we merge the dictionary from b into the
                    dictionary from a using the given strategy."""
                    a_has_changed = a_has_changed or self.__merge_trees(
                        a[key], b[key], merge_strategy
                    )
                    if (
                        a_has_changed
                        and isinstance(a, ParamStore)
                        and isinstance(b, ParamStore)
                    ):
                        """if the dictionary in a has been changed, this was done
                        in-place. We therefore need to mark the Param key as dirty. We
                        only do this at the very top of the recursion - when a and b
                        are the ParamStores being merged"""
                        self.__dirty_keys.add(key)
                elif a[key] == b[key]:
                    pass  # same leaf values, nothing to do
                else:  # diff leave values => conflict:
                    if merge_strategy == MergeStrategy.OURS:
                        pass  # a takes precedence, ignore b
                    elif merge_strategy == MergeStrategy.THEIRS:
                        a[key] = b[key]  # b takes precedence, overwrite a
                        a_has_changed = True
                    else:
                        raise NotImplementedError(
                            f"MergeStrategy '{merge_strategy}' is not implemented"
                        )
            else:  # key from b is not in a:
                a[key] = b[key]  # "copy" from b to a
                a_has_changed = True
        return a_has_changed

    """ Diff """

    def diff(
        self, old_commit_id: Optional[str] = None, new_commit_id: Optional[str] = None
    ) -> Dict[str, Dict]:
        """Shows the difference in Param values between two commits.

        :param old_commit_id: The id of the first ("older") commit to compare. If
            None, or not specified, defaults to the latest commit id.
        :param new_commit_id: The id of the second  ("newer") commit to compare. If
            None, or not specified, defaults to the current state of the store (incl.
             "dirty" values)
        :return: A dictionary where keys are the keys of params whose values have
            changed. Dictionary values indicate the `old_value` of the param and the
            `new_value` of the param. A new param will only show the `new_value`. A
            deleted param will only show the `old_value`.
            Example: {"foo": {"old_value": "bar", "new_value": "baz"}}
        """
        with self.__lock:

            # get OLD params to diff:
            if old_commit_id:
                old_commit = self.__persistence.get_commit(old_commit_id)
            else:  # default to latest commit
                old_commit = self.__persistence.get_latest_commit()
            old_params = old_commit.params if old_commit else {}

            # get NEW params to diff:
            if new_commit_id:
                new_commit = self.__persistence.get_commit(new_commit_id)
                new_params = new_commit.params if new_commit else {}
            else:  # default to dirty params
                new_params = self.__params

            return self.__diff(old_params, new_params)

    @staticmethod
    def __diff(old: MutableMapping, new: MutableMapping) -> Dict[str, Dict]:
        diff = dict()
        for key in new.keys():
            if key in old.keys():
                old_value = old[key].value
                new_value = new[key].value
                if old_value != new_value:  # different values
                    diff[key] = dict(old_value=old_value, new_value=new_value)
            else:
                diff[key] = dict(new_value=new[key].value)  # added
        for key in old.keys():
            if key not in new.keys():
                diff[key] = dict(old_value=old[key].value)  # deleted
        return diff

    @staticmethod
    def __safe_get_value_from_params(params: Dict[str, Param], key: str) -> Optional:
        if key in params:
            return params[key].value
        else:
            return None

    def list_values(self, key: str) -> pd.DataFrame:
        """
        Lists all the values of a given key taken from commit history,
        sorted by date ascending

        :param key: the key for which to list values
        :returns: a list of tuples where the values are, in order:
            - the value of the key
            - time of commit
            - commit_id
            - label assigned to commit
        """
        with self.__lock:
            values = []
            commits = self.__persistence.search_commits(key=key)
            commits.sort(key=lambda c: c.timestamp)
            for commit in commits:
                value = (
                    commit.params[key].value,
                    commit.local_timestamp,
                    commit.id,
                    commit.label,
                )
                values.append(value)
            if self.is_dirty and key in self.__params.keys():
                values.append((self[key], None, None, None))
            df = pd.DataFrame(values)
            if not df.empty:
                df.columns = ["value", "time", "commit_id", "label"]
            return df

    """ Tags """

    def add_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if key not in self.__params.keys():
                raise KeyError(f"key '{key}' is not in store")
            if tag not in self.__tags:
                self.__tags[tag] = []
            self.__tags[tag].append(key)

    def remove_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if tag not in self.__tags:
                return
            if key not in self.__tags[tag]:
                return
            self.__tags[tag].remove(key)

    def list_keys_for_tag(self, tag: str) -> List[str]:
        with self.__lock:
            if tag not in self.__tags:
                return []
            else:
                return self.__tags[tag]

    def list_tags_for_key(self, key: str):
        tags_for_key = []
        for item in self.__tags.items():
            if key in item[1]:
                tags_for_key.append(item[0])
        return tags_for_key

    """ Temporary State """

    def save_temp(self) -> None:
        """
        Saves the state of params to a temporary location
        """
        with self.__lock:
            commit = Commit(
                params=self.__params,
                tags=self.__tags,
            )
            self.__persistence.save_temp_commit(commit)

    def load_temp(self) -> None:
        """
        Overwrites the current state of params with data loaded from the temporary
        location
        """
        with self.__lock:
            commit = self.__persistence.load_temp_commit()
            self.__params.clear()
            self.__params.update(commit.params)
            self.__tags.clear()
            self.__tags.update(commit.tags)


""" Static helper methods """


def _map_dict(f, d: Dict) -> Dict:
    """Applies function f to all values in dict d recursively"""
    values_dict = dict()
    for item in d.items():
        k = item[0]
        v = item[1]
        if not isinstance(v, Param) and isinstance(v, dict):
            values_dict[k] = _map_dict(f, v)
        else:
            values_dict[k] = f(v)
    return values_dict


def _extract_param_values(d: Dict) -> Dict:
    return _map_dict(lambda x: x.value, d)



"""initial migration

Revision ID: 000c6a88457f
Revises:
Create Date: 2022-08-23 15:01:32.181858+00:00

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "000c6a88457f"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "commit",
        sa.Column("id", postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column(
            "timestamp",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("label", sa.String(length=256), nullable=True),
        sa.Column("params", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("tags", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "temp",
        sa.Column("id", postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column(
            "timestamp",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column("label", sa.String(length=256), nullable=True),
        sa.Column("params", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("tags", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("temp")
    op.drop_table("commit")
    # ### end Alembic commands ###



"""Fixing Param qualified names as they appear in the ParamStore JSON file

Revision ID: 09f3b5a1689c
Revises: 273a9fae6206
Create Date: 2022-06-28 12:13:39.743933+00:00

"""
import shutil

from entropylab.logger import logger
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.params.persistence.migrations import fix_param_qualified_name
from entropylab.pipeline.params.persistence.tinydb.tinydbpersistence import set_version
from entropylab.pipeline.results_backend.sqlalchemy.alembic.alembic_util import (
    AlembicUtil,
)

# revision identifiers, used by Alembic.

revision = "09f3b5a1689c"
down_revision = "273a9fae6206"
branch_labels = None
depends_on = None


def upgrade():
    path = str(AlembicUtil.get_param_store_file_path())
    logger.debug(f"Starting to fix Param qualified name in ParamStore file [{path}]")
    try:
        fix_param_qualified_name(path, revision)
        set_version(path, "0.2", revision)
    except EntropyError as ee:
        logger.warning(str(ee))
    logger.debug("Done fixing Param qualified name.")


def downgrade():
    path = AlembicUtil.get_param_store_file_path()
    backup_json_file_path = path.replace(".json", ".json.{revision}.bak")
    downgraded_json_file_path = path.replace(".json", ".json.{revision}.downgraded")
    logger.debug(
        f"Attempting to downgrade ParamStore file {path} from revision "
        f"{revision} to {down_revision} using backup file {backup_json_file_path}"
    )
    shutil.move(path, downgraded_json_file_path)
    shutil.move(backup_json_file_path, path)
    logger.debug(f"Done downgrading from {revision} to {down_revision}")



"""Deleting results and metadata from sqlite

Revision ID: 7fa75ca1263f
Revises: 06140c96c8c4
Create Date: 2022-06-16 13:26:06.576200+00:00

"""
import re
import shutil

from alembic import op

# revision identifiers, used by Alembic.
from entropylab.logger import logger

revision = "7fa75ca1263f"
down_revision = "06140c96c8c4"
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    db_path = conn.engine.url.database
    if db_path == ":memory:":
        return
    backup_path = get_backup_path(db_path)

    logger.debug(f"Backing up database to [{backup_path}] ")
    shutil.copy2(db_path, backup_path)

    logger.debug("Deleting Result records that are saved in HDF5")
    op.execute("DELETE FROM Results WHERE saved_in_hdf5=1;")

    logger.debug("Deleting ExperimentMetadata records that are saved in HDF5")
    op.execute("DELETE FROM ExperimentMetadata WHERE saved_in_hdf5=1;")

    logger.debug("Vacuuming DB to reduce its size")
    with op.get_context().autocommit_block():
        op.execute("VACUUM")


def downgrade():
    conn = op.get_bind()
    db_path = conn.engine.url.database
    if db_path == ":memory:":
        return
    backup_path = get_backup_path(db_path)
    downgraded_db_path = get_downgraded_db_path(db_path)
    logger.debug(
        f"Attempting to downgrade Entropy DB {db_path} "
        f"using backup file {backup_path}"
    )
    shutil.move(db_path, downgraded_db_path)
    shutil.move(backup_path, db_path)


def get_backup_path(db_path: str) -> str:
    return re.sub(".db$", f".bak.{revision}.db", db_path)


def get_downgraded_db_path(db_path: str) -> str:
    return re.sub(".db$", f".downgraded.{revision}.db", db_path)



import os
import uuid
from pathlib import Path
from typing import Optional, Set, List

import jsonpickle
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Connection
from sqlalchemy.orm import sessionmaker

from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.params.persistence.persistence import Persistence, Commit
from entropylab.pipeline.params.persistence.sqlalchemy.model import (
    CommitTable,
    TempTable,
)

TEMP_COMMIT_ID = "00000000-0000-0000-0000-000000000000"


class SqlAlchemyPersistence(Persistence):
    def __init__(self, url: Optional[str] = None):
        self.engine = create_engine(
            url,
            json_serializer=jsonpickle.encode,
            json_deserializer=jsonpickle.decode,
        )
        self.__session_maker = sessionmaker(bind=self.engine)
        if self._db_is_empty():
            self.__init()

    def __init(self):
        with self.engine.connect() as connection:
            alembic_cfg = self.__alembic_build_config(connection)
            command.upgrade(alembic_cfg, "head")

    def __alembic_build_config(self, connection: Connection) -> Config:
        config_location = self._abs_path_to("alembic.ini")
        script_location = self._abs_path_to("alembic")
        alembic_cfg = Config(config_location)
        alembic_cfg.set_main_option("script_location", script_location)
        """
        Modified by @urig to share a single engine across multiple (typically in-memory)
        connections, based on this cookbook recipe:
        https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
        """
        alembic_cfg.set_main_option("sqlalchemy.url", "sqlite:///:memory:")
        alembic_cfg.attributes["connection"] = connection  # overrides dummy url above
        return alembic_cfg

    @staticmethod
    def _abs_path_to(rel_path: str) -> str:
        source_path = Path(__file__).resolve()
        source_dir = source_path.parent
        return os.path.join(source_dir, rel_path)

    def close(self):
        self.__session_maker.close_all()

    def get_commit(
        self, commit_id: Optional[str] = None, commit_num: Optional[int] = None
    ):
        if commit_id:
            with self.__session_maker() as session:
                commit = (
                    session.query(CommitTable)
                    .filter(CommitTable.id == commit_id)
                    .one_or_none()
                )
                if commit:
                    return commit
                else:
                    raise EntropyError(f"Commit with id '{commit_id}' not found")
        elif commit_num:
            with self.__session_maker() as session:
                commit = (
                    session.query(CommitTable)
                    .order_by(CommitTable.timestamp.asc())
                    .offset(commit_num - 1)
                    .limit(1)
                    .one_or_none()
                )
                if commit:
                    return commit
                else:
                    raise EntropyError(f"Commit with id '{commit_id}' not found")

        else:
            return self.get_latest_commit()

    def get_latest_commit(self):
        with self.__session_maker() as session:
            commit = (
                session.query(CommitTable)
                .order_by(CommitTable.timestamp.desc())
                .first()
            )
            return commit

    def commit(
        self,
        commit: Commit,
        dirty_keys: Optional[Set[str]] = None,
    ) -> str:
        commit.id = self.__generate_commit_id()
        # TODO: Perhaps create the timestamp here?
        self.stamp_dirty_params_with_commit(commit, dirty_keys)
        commit_table = CommitTable()
        commit_table.id = commit.id
        commit_table.timestamp = commit.timestamp
        commit_table.label = commit.label
        commit_table.params = commit.params
        commit_table.tags = commit.tags
        with self.__session_maker() as session:
            session.add(commit_table)
            session.commit()
            return commit_table.id

    @staticmethod
    def __generate_commit_id() -> str:
        return str(uuid.uuid4())

    def search_commits(
        self, label: Optional[str] = None, key: Optional[str] = None
    ) -> List[Commit]:
        with self.__session_maker() as session:
            commits = session.query(CommitTable)
            if label:
                commits = commits.filter(CommitTable.label == label)
            if key:
                commits = commits.filter(CommitTable.params.contains(key))
            return commits.all()

    def save_temp_commit(self, commit: Commit) -> None:
        with self.__session_maker() as session:
            temp = session.get(TempTable, TEMP_COMMIT_ID)
            if temp:
                temp.timestamp = commit.timestamp
                temp.label = commit.label
                temp.params = commit.params
                temp.tags = commit.tags
                session.commit()
            else:
                temp = TempTable()
                temp.id = TEMP_COMMIT_ID
                temp.timestamp = commit.timestamp
                temp.label = commit.label
                temp.params = commit.params
                temp.tags = commit.tags
                session.add(temp)
                session.commit()

    def load_temp_commit(self) -> Commit:
        with self.__session_maker() as session:
            temp = session.get(TempTable, TEMP_COMMIT_ID)
            if not temp:
                raise EntropyError(
                    "Temp is empty. Use save_temp_commit() before using "
                    "load_temp_commit() "
                )
            else:
                commit = Commit(
                    id=temp.id,
                    timestamp=temp.timestamp,
                    label=temp.label,
                    params=temp.params,
                    tags=temp.tags,
                )
            return commit

    def _db_is_empty(self) -> bool:
        with self.__session_maker() as session:
            result = session.execute(
                text("SELECT sql FROM sqlite_master WHERE type = 'table'")
            )
            return len(result.fetchall()) == 0



import os
import shutil
from pathlib import Path
from typing import TypeVar, Type, Tuple

import sqlalchemy.engine
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from entropylab.logger import logger
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.results_backend.sqlalchemy.alembic.alembic_util import (
    AlembicUtil,
)
from entropylab.pipeline.results_backend.sqlalchemy.model import (
    Base,
    ResultTable,
    MetadataTable,
)
from entropylab.pipeline.results_backend.sqlalchemy.project import (
    project_name,
    project_path,
)
from entropylab.pipeline.results_backend.sqlalchemy.storage import (
    HDF5Storage,
    EntityType,
)

T = TypeVar("T", bound=Base)

_SQL_ALCHEMY_MEMORY = ":memory:"
_ENTROPY_DIRNAME = ".entropy"
_DB_FILENAME = "entropy.db"
_HDF5_FILENAME = "entropy.hdf5"
_HDF5_DIRNAME = "hdf5"


class _DbInitializer:
    def __init__(self, path: str, echo=False):
        """
        :param path: path to directory containing Entropy project
        :param echo: if True, the database engine will log all statements
        """
        self._validate_path(path)

        if path is None or path == _SQL_ALCHEMY_MEMORY:
            logger.debug("_DbInitializer is in in-memory mode")
            self._storage = HDF5Storage()
            self._engine = create_engine("sqlite:///" + _SQL_ALCHEMY_MEMORY, echo=echo)
            self._alembic_util = AlembicUtil(self._engine)
        else:
            logger.debug("_DbInitializer is in project directory mode")
            creating_new = not os.path.isdir(path)
            entropy_dir_path = os.path.join(path, _ENTROPY_DIRNAME)
            os.makedirs(entropy_dir_path, exist_ok=True)
            logger.debug(f"Entropy directory is at: {entropy_dir_path}")

            db_file_path = os.path.join(entropy_dir_path, _DB_FILENAME)
            logger.debug(f"DB file is at: {db_file_path}")

            hdf5_dir_path = os.path.join(entropy_dir_path, _HDF5_DIRNAME)
            logger.debug(f"hdf5 directory is at: {hdf5_dir_path}")

            self._engine = create_engine("sqlite:///" + db_file_path, echo=echo)
            self._storage = HDF5Storage(hdf5_dir_path)
            self._alembic_util = AlembicUtil(self._engine)
            if creating_new:
                self._print_project_created(path)

    def init_db(self) -> Tuple[sqlalchemy.engine.Engine, HDF5Storage]:
        """If the database is empty, initializes it with the most up-to-date schema.
        If the database is not empty, ensures that is up-to-date."""
        if self._db_is_empty():
            Base.metadata.create_all(self._engine)
            self._alembic_util.stamp_head()
        else:
            if not self._alembic_util.db_is_up_to_date():
                path = str(self._engine.url)
                raise EntropyError(
                    f"The database at {path} is not up-to-date. Update the database "
                    "using the Entropy CLI command `entropy upgrade`. "
                    "* Before upgrading be sure to back up your database to a safe "
                    "place *."
                )
        return self._engine, self._storage

    @staticmethod
    def _print_project_created(path):
        print(
            f"New Entropy project '{project_name(path)}' created at "
            f"'{project_path(path)}'"
        )

    @staticmethod
    def _validate_path(path):
        if path is None or path == _SQL_ALCHEMY_MEMORY:
            return
        if path is not None and Path(path).suffix == ".db":
            logger.error(
                "_DbInitializer given path to a sqlite database. This is deprecated."
            )
            raise EntropyError(
                "Providing the SqlAlchemyDB() constructor with a path to a sqlite "
                "database is deprecated. You should instead provide the path to a "
                "directory containing an Entropy project.\n"
                "To upgrade your existing sqlite database file to an Entropy project "
                "please use the Entropy CLI command: `entropy upgrade`.\n"
                "* Before upgrading be sure to back up your database to a safe place *."
            )
        if path is not None and os.path.isfile(path):
            logger.error(
                "_DbInitializer provided with path to a file, not a directory."
            )
            raise RuntimeError(
                "SqlAlchemyDB() constructor provided with a path to a file but "
                "expects the path to an Entropy project directory"
            )

    def _db_is_empty(self) -> bool:
        cursor = self._engine.execute(
            "SELECT sql FROM sqlite_master WHERE type = 'table'"
        )
        return len(cursor.fetchall()) == 0


class _DbUpgrader:
    def __init__(self, path: str, echo=False) -> None:
        self._path = path
        self._echo = echo
        self._engine = None
        self._storage = None
        self._alembic_util = None

    def upgrade_db(self) -> None:
        old_global_hdf5_file_path = None
        if self._path is None or self._path == _SQL_ALCHEMY_MEMORY:
            logger.debug("_DbUpgrader is in in-memory mode")
            self._storage = HDF5Storage()
            self._engine = create_engine(
                "sqlite:///" + _SQL_ALCHEMY_MEMORY, echo=self._echo
            )
        else:
            logger.debug("_DbUpgrader is in project directory mode")
            if not os.path.exists(self._path):
                raise EntropyError(f"No Entropy project exists at '{self._path}'")
            if self._path_is_to_db():
                self._convert_to_project()
            entropy_dir_path = os.path.join(self._path, _ENTROPY_DIRNAME)
            db_file_path = os.path.join(entropy_dir_path, _DB_FILENAME)
            hdf5_dir_path = os.path.join(entropy_dir_path, _HDF5_DIRNAME)
            old_global_hdf5_file_path = os.path.join(entropy_dir_path, _HDF5_FILENAME)
            self._engine = create_engine("sqlite:///" + db_file_path, echo=self._echo)
            self._storage = HDF5Storage(hdf5_dir_path)
        self._alembic_util = AlembicUtil(self._engine)
        self._alembic_util.upgrade()
        if old_global_hdf5_file_path and os.path.isfile(old_global_hdf5_file_path):
            # old, global hdf5 file exists so migrate from it to new "per experiment"
            # hdf5 files
            self._storage.migrate_from_per_project_hdf5_to_per_experiment_hdf5_files(
                old_global_hdf5_file_path
            )
        else:
            # old, global hdf5 file does not exist so migrate directly from DB to new
            # "per experiment" hdf5 files
            self._migrate_results_from_db_to_hdf5()
            self._migrate_metadata_from_db_to_hdf5()
        self.log_success()

    def log_success(self):
        at = f"at [{self._path}] " if self._path != _SQL_ALCHEMY_MEMORY else ""
        logger.info(f"The Entropy project {at}was updated successfully.")

    def _path_doesnt_exist(self):
        return not os.path.isdir(self._path) and not os.path.isfile(self._path)

    def _path_is_to_db(self):
        return (
            self._path is not None
            and (Path(self._path).suffix == ".db")
            or os.path.isfile(self._path)
        )

    def _convert_to_project(self):
        # old
        old_db_file_path = self._path
        old_hdf5_file_path = old_db_file_path.replace(".db", ".hdf5")
        # new
        new_project_dir_path = os.path.splitext(old_db_file_path)[0]
        new_entropy_dir_path = os.path.join(new_project_dir_path, _ENTROPY_DIRNAME)
        new_db_file_path = os.path.join(new_entropy_dir_path, _DB_FILENAME)
        new_hdf5_file_path = os.path.join(new_entropy_dir_path, _HDF5_FILENAME)
        # convert
        os.makedirs(new_entropy_dir_path, exist_ok=True)
        shutil.move(old_db_file_path, new_db_file_path)
        if os.path.exists(old_hdf5_file_path):
            shutil.move(old_hdf5_file_path, new_hdf5_file_path)
        self._path = new_project_dir_path
        print(
            f"Converted db file at {old_db_file_path} to project directory "
            f"at {new_project_dir_path}"
        )

    def _migrate_results_from_db_to_hdf5(self) -> None:
        self._migrate_rows_from_db_to_hdf5(EntityType.RESULT, ResultTable)

    def _migrate_metadata_from_db_to_hdf5(self) -> None:
        self._migrate_rows_from_db_to_hdf5(EntityType.METADATA, MetadataTable)

    def _migrate_rows_from_db_to_hdf5(self, entity_type: EntityType, table: Type[T]):
        logger.debug(f"Migrating {entity_type.name} rows from sqlite to hdf5")
        session_maker = sessionmaker(bind=self._engine)
        with session_maker() as session:
            rows = session.query(table).filter(table.saved_in_hdf5.is_(False)).all()
            if len(rows) == 0:
                logger.debug(f"No {entity_type.name} rows need migrating. Done")
            else:
                logger.debug(f"Found {len(rows)} {entity_type.name} rows to migrate")
                self._storage.migrate_rows(entity_type, rows)
                logger.debug(f"Migrated {len(rows)} {entity_type.name} to hdf5")
                for row in rows:
                    row.saved_in_hdf5 = True
                session.commit()
                logger.debug(
                    f"Marked all {entity_type.name} rows in sqlite as `saved_in_hdf5`."
                    f" Done"
                )



from abc import abstractmethod, ABC
from typing import Callable, Dict

from munch import Munch
from qm.QuantumMachinesManager import QuantumMachinesManager
from qualang_tools.config import ConfigBuilder
from qualang_tools.config import ConfigurationError
from qualang_tools.config.parameters import ConfigVars

from entropylab import ParamStore


class QuAMManager(ABC):
    def __init__(self, path: str, host=None, port=None, **kwargs):
        """
        QuAMManager
        :param host: Host where to find the QM orchestrator. If ``None``,
            local settings are used.
        :type host: str
        :param port: Port where to find the QM orchestrator. If None, local settings
         are used):
        :type port: int
        :param path: Path of the entropy DB
        :type path: str
        """
        self.path = path
        self.param_store = ParamStore(path)
        self._config_builder = ConfigBuilder()
        self._config_vars = ConfigVars()
        self.config = {}
        self.host = host
        self.port = port
        self.qmm = QuantumMachinesManager(host=host, port=port, **kwargs)

    def generate_config(self):
        """Returns the QUA configuration"""
        self._config_builder = ConfigBuilder()
        self.prepare_config(self._config_builder)
        self._set_config_vars()
        self.config = self._config_builder.build()
        return self.config

    @property
    def config_builder(self):
        return self._config_builder

    @abstractmethod
    def prepare_config(self, cb: ConfigBuilder) -> None:
        """method to prepare the ConfigBuilder, needs to be implemented by the QuAMManager"""
        pass

    def parameter(self, var: str, setter: Callable = None):
        """Returns a parameter with the given name and setter, all parameters are saved
        in the ParamStore
        :return: a parameter
        :rtype: Parameter
        """
        if var not in self.param_store.keys():
            self.param_store[var] = None
        return self._config_vars.parameter(var, setter=setter)

    def _set_config_vars(self):
        """Sets the parameter values according to the key, value pairs in the param store"""
        _dict = {}
        for k in self.param_store.keys():
            if not callable(self._config_vars.parameter(k)._value):
                if self.param_store[k] is None:
                    raise ConfigurationError("Set parameter {}".format(k))
                _dict[k] = self.param_store[k]
        self._config_vars.set(**_dict)

    @property
    def elements(self):
        """Returns a Munch instance of element keys available in the QUA configuration"""
        return get_key("elements", self.config)

    @property
    def pulses(self):
        """Returns a Munch instance of pulses available in the QUA configuration"""
        return get_key("pulses", self.config)

    @property
    def integration_weights(self):
        """Returns a Munch instance of integration weights available in the QUA configuration"""
        return get_key("integration_weights", self.config)

    def open_qm(self):
        return self.qmm.open_qm(self.generate_config())


def dict_keys_to_munch(d: Dict):
    elms = Munch()
    for k in d.keys():
        elms[k] = k
    return elms


def get_key(k: str, d: dict):
    if k in d:
        return dict_keys_to_munch(d[k])
    else:
        raise KeyError(
            "{} doesn't exist in the QUA configuration, probably the \
                config wasn't generated".format(
                k
            )
        )
